/* FILE: tclcmds.cc                 -*-Mode: c++-*-
 *
 * Commands added to Tcl/Tk interpreter (& supporting routines)
 *
 */

#include <math.h>	/* rint, ... */
#include <tcl.h>	/* TCL_OK, ... */
#include <tk.h>		/* Tk_Window, ... */

#include "mesh.h"       /* Abstract mesh type and descendents */
#include "misc.h"       /* Utility functions */
#include "mmdisp.h"	/* MMDISP_MAJOR_VERSION, ... */
#include "tclcmds.h"	/* protoypes */
#include "display.h"	/* DisplayFrame ... */
#include "misc.h"	/* PlainError, ... */
#include "vecfile.h"    /* Vector field file input routines */

// Load icon bitmap and mask (except under Windows).
// These files must define
//      omficon_width omficon_height omficon_bits
//  and
//      omfmask_width omfmask_height omfmask_bits
// respectively.
#if SYSTEM_TYPE!=WINDOWS
#include "omficon.xbm"
#include "omfmask.xbm"

static Pixmap omfIconBitmap;
static Pixmap omfIconMask;
#endif

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

static Tk_Window myTkWin;
static Display *myDisplay;

static char *ExeName=NULL;
static char *FirstFileName=NULL;

static Mesh *myMesh;
static DisplayFrame myFrame;

static INT4m ServerFlag=0;

static CHAR buf[SCRATCH_BUF_SIZE];
static CHAR buf2[SCRATCH_BUF_SIZE];

void Mmd_SaveServerFlag(INT2m value)
{
  // Save the value in a static variable
  ServerFlag = value;
}

int TclCommandsPreInit(char* runname,char* inputfile)
{
  // Initialize myMesh with an empty mesh
  myMesh=new EmptyMesh();

  // Save executable name
  ExeName=new char[strlen(runname)+1];
  strcpy(ExeName,runname);
  // Save first image file name
  if(inputfile!=NULL) {
    FirstFileName=new char[strlen(inputfile)+1];
    strcpy(FirstFileName,inputfile);
  }
  return 0;
}

void TclCommandsCleanup(ClientData)
{
#if SYSTEM_TYPE!=WINDOWS
  // I don't know if the following Tk_FreeBitmap are
  // needed or valid, but I'll keep them here until
  // they cause problems.
  Tk_FreeBitmap(myDisplay,omfIconBitmap);
  Tk_FreeBitmap(myDisplay,omfIconMask);
#endif

  // Free C++ allocated memory
  if(ExeName!=NULL) delete[] ExeName;  // Safety-check on Null
  if(FirstFileName!=NULL) delete[] FirstFileName;
  if(myMesh!=NULL) delete myMesh;
#if (DEBUGLVL>1)
  PlainWarning("TclCommandsCleanup routine finished.");
#endif
}

static REAL4m CoordsToAngle(CoordinateSystem coords)
{
  REAL4m rotang=0.;
  switch(coords) {
  case CS_CALCULATION_STANDARD: // Safety
  case CS_DISPLAY_STANDARD:  rotang=0.;   break;
  case CS_DISPLAY_ROT_90:    rotang=90.;  break;
  case CS_DISPLAY_ROT_180:   rotang=180.; break;
  case CS_DISPLAY_ROT_270:   rotang=270.; break;
  case CS_ILLEGAL:
  default:
    PlainError(1,"Unrecognized or illegal coordinate system (%d)"
	       " detected in CoordsToAngle() (File tkcmds.cc)\n",
	       int(coords));
    break; // Shouldn't get to here
  }
  return rotang;
}

static CoordinateSystem AngleToCoords(REAL4m angle_in_degrees)
{ // Rounds to nearest supported coordinate system rotation
  CoordinateSystem coords=CS_ILLEGAL;
  int quad=int(rint(fmod(angle_in_degrees,360.)/90.));
  if(quad<0) quad+=4;  if(quad>3) quad-=4;
  switch(quad) {
  case 0:  coords=CS_DISPLAY_STANDARD;  break;
  case 1:  coords=CS_DISPLAY_ROT_90;    break;
  case 2:  coords=CS_DISPLAY_ROT_180;   break;
  case 3:  coords=CS_DISPLAY_ROT_270;   break;
  default:
    PlainError(1,"Programming error: Illegal coordinate system (quad=%d)"
	       " occurred in AngleToCoords() (File tkcmds.cc)\n",quad);
    break; // Shouldn't get to here
  }
  return coords;
}

int GetFirstFileName(ClientData,Tcl_Interp *interp,int argc,char **)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetFirstFileName must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  if(FirstFileName!=NULL) strncpy(buf,FirstFileName,sizeof(buf));
  else                    buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshType(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshType must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *name=myMesh->GetMeshType();
  if(name!=NULL) strncpy(buf,name,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshName(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshName must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *name=myMesh->GetName();
  if(name!=NULL) strncpy(buf,name,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshTitle(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshTitle must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *title=myMesh->GetTitle();
  if(title!=NULL) strncpy(buf,title,sizeof(buf));
  else            buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshDescription(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshDesc must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  const char *desc=myMesh->GetDescription();
  if(desc!=NULL) strncpy(buf,desc,sizeof(buf));
  else           buf[0]='\0';
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int GetMeshIncrement(ClientData,Tcl_Interp *interp,int argc,char**)
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    sprintf(buf,"GetMeshIncrement must be called with no arguments"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  sprintf(buf,"%f",myMesh->GetSubsampleGrit());
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

int ChangeMesh(ClientData,Tcl_Interp *interp,int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<5 || argc>8) {
    sprintf(buf,"ChangeMesh must be called with 4-7 arguments:"
	    " mesh_filename"
	    " frame_width frame_height rotation_degrees"
	    " [new_sample_step [new_arrowmag [new_scale]]]"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  strcpy(buf,"OMF_PushCursor ; OMF_SoftUpdate");
  Tcl_Eval(interp,buf);    // Save current cursor & display "watch" cursor

  delete myMesh;
  if(argv[1]==NULL || argv[1][0]=='\0') { // Null request
    myMesh=new EmptyMesh();
  }
  else {
    VectorFieldFileInput *vffreader
      = VectorFieldFileInput::NewReader(argv[1]);
    if(vffreader==NULL) { // Unknown file type
      myMesh=new EmptyMesh();
    }
    else { // Reader found for file argv[1]
      myMesh=vffreader->NewMesh();
      delete vffreader;
      if(myMesh==NULL) myMesh=new EmptyMesh(); // Safety
    }
  }

  myFrame.SetMesh(myMesh);
  REAL4m width=Omf_atof(argv[2]);
  REAL4m height=Omf_atof(argv[3]);
  CoordinateSystem coords=AngleToCoords(Omf_atof(argv[4]));
  REAL4m new_sample_step=-1.;
  if(argc>5) {
    new_sample_step=Omf_atof(argv[5]);
  }
  REAL4m new_arrowmag=-1.;
  if(argc>6) {
    new_arrowmag=Omf_atof(argv[6]);
  }
  REAL4m new_scale=-1.;
  if(argc>7) {
    new_scale=Omf_atof(argv[7]);
  }

  myFrame.SetupFrame(width,height,coords,new_sample_step,
		     new_arrowmag,new_scale);

  strcpy(buf,"OMF_PopCursor");
  Tcl_Eval(interp,buf);    // Retrieve original cursor

  return TCL_OK;
}

int DrawFrame(ClientData,Tcl_Interp *interp,int argc,char** argv)
{
  int error_code;
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"DrawFrame must be called with 1 argument: canvas_name"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  strcpy(buf,"OMF_PushCursor ; OMF_SoftUpdate");
  Tcl_Eval(interp,buf);    // Save current cursor & display "watch" cursor

  sprintf(buf,"%s delete all ;# update idletasks",argv[1]);
  /// Uncomment the "update idletasks" command to get a white
  /// screen between frames
  Tcl_Eval(interp,buf); // Clear canvas
  error_code=myFrame.Show(argv[1]); // Redraw canvas

  sprintf(buf,"global DrawFrameCount; incr DrawFrameCount");
  Tcl_Eval(interp,buf); // Increment DrawFrameCount variable

  strcpy(buf,"OMF_PopCursor");
  Tcl_Eval(interp,buf);    // Retrieve original cursor

  if(error_code!=TCL_OK) {
    sprintf(buf,"Error from myFrame.Show(%s)",argv[1]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return error_code;
  }
  return TCL_OK;
}

int GetFrameRotation(ClientData,Tcl_Interp *interp,
		     int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"GetFrameRotation must be called with"
	    " 1 argument: name of Tk's Frame Rotation variable"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  INT4m rotang=int(rint(CoordsToAngle(myFrame.GetCoordinates())));

  sprintf(buf,"%d",rotang);
  Tcl_SetVar(interp,argv[1],buf,0);

  return TCL_OK;
}

int SetFrameRotation(ClientData,Tcl_Interp *interp,
		     int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"SetFrameRotation must be called with"
	    " 1 argument: new_rotation_angle"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  strcpy(buf,"OMF_PushCursor ; OMF_SoftUpdate");
  Tcl_Eval(interp,buf);    // Save current cursor & display "watch" cursor

  myFrame.ChangeCoordinates(AngleToCoords(Omf_atof(argv[1])));

  strcpy(buf,"OMF_PopCursor");
  Tcl_Eval(interp,buf);    // Retrieve original cursor

  return TCL_OK;
}

int GetFrameSubsampling(ClientData,Tcl_Interp *interp,
			int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"GetFrameSubsampling must be called with"
	    " 1 argument: name of Tk's Subsample variable"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  sprintf(buf,"%8g",myFrame.GetSampleStepSize());
  Tcl_SetVar(interp,argv[1],buf,0);

  return TCL_OK;
}

int SetFrameSubsampling(ClientData,Tcl_Interp *interp,
			int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"SetFrameSubsampling must be called with"
	    " 1 argument: new_sample_step"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  REAL4 new_sample_step=Omf_atof(argv[1]);
  if(new_sample_step<0.0) {
    sprintf(buf,"Sample step must be >= 0.0");
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  strcpy(buf,"OMF_PushCursor ; OMF_SoftUpdate");
  Tcl_Eval(interp,buf);    // Save current cursor & display "watch" cursor

  // Change stepsize in display list
  if(myFrame.SetSampleStepSize(new_sample_step)!=0) return TCL_ERROR;

  strcpy(buf,"OMF_PopCursor");
  Tcl_Eval(interp,buf);    // Retrieve original cursor

  return TCL_OK;
}

int GetFrameSize(ClientData,Tcl_Interp *interp,
		 int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=3) {
    sprintf(buf,"GetFrameSize should be called with 2 arguments:"
	    " name of framewidth & frameheight variables."
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  BoundingBox bb=myFrame.GetDisplayBox();

  sprintf(buf,"%4d",int(rint(bb.GetWidth())));
  Tcl_SetVar(interp,argv[1],buf,0);

  sprintf(buf,"%4d",int(rint(bb.GetHeight())));
  Tcl_SetVar(interp,argv[2],buf,0);

  return TCL_OK;
}

int SetFrameSize(ClientData,Tcl_Interp *interp,
		 int argc,char** argv)
{ // Display size *request*!
  Tcl_ResetResult(interp);
  if(argc!=3) {
    sprintf(buf,"SetFrameSize should be called with 2 arguments:"
	    " ImageWidth & ImageHeight."
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  BoundingBox bb=myFrame.GetDisplayBox();
  REAL4m oldwidth=bb.GetWidth();
  REAL4m oldheight=bb.GetHeight();
  REAL4m newwidth=Omf_atof(argv[1]);
  REAL4m newheight=Omf_atof(argv[2]);
  if(oldwidth==newwidth && oldheight==newheight)
    return TCL_OK;  // No change

  strcpy(buf,"OMF_PushCursor ; OMF_SoftUpdate");
  Tcl_Eval(interp,buf);    // Save current cursor & display "watch" cursor

  Vec3f minpt; bb.GetMinPt(minpt);
  Vec3f maxpt=minpt;
  maxpt.x+=newwidth; maxpt.y+=newheight; maxpt.z+=-1;
  bb.Set(minpt,maxpt);
  myFrame.SetDisplayBox(bb);

  strcpy(buf,"OMF_PopCursor");
  Tcl_Eval(interp,buf);    // Retrieve original cursor

  return TCL_OK;
}

int GetFrameScale(ClientData,Tcl_Interp *interp,
		  int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"GetFrameScale must be called with"
	    " 1 argument: name of Tk's FrameScale variable"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  sprintf(buf,"%8g",myFrame.GetScale());
  Tcl_SetVar(interp,argv[1],buf,0);

  return TCL_OK;
}

int SetFrameScale(ClientData,Tcl_Interp *interp,
		  int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"SetFrameScale must be called with"
	    " 1 argument: new_scale"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  REAL4 new_scale=Omf_atof(argv[1]);
  if(new_scale<0.0) {
    sprintf(buf,"Scale must be >= 0.0");
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  strcpy(buf,"OMF_PushCursor ; OMF_SoftUpdate");
  Tcl_Eval(interp,buf);    // Save current cursor & display "watch" cursor

  // Change scale in display list
  if(myFrame.SetScale(new_scale)!=0) return TCL_ERROR;

  strcpy(buf,"OMF_PopCursor");
  Tcl_Eval(interp,buf);    // Retrieve original cursor

  return TCL_OK;
}

int GetFrameArrowMag(ClientData,Tcl_Interp *interp,
		  int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"GetFrameArrowMag must be called with"
	    " 1 argument: name of Tk's FrameArrowMag variable"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  sprintf(buf,"%8g",myFrame.GetArrowMag());
  Tcl_SetVar(interp,argv[1],buf,0);

  return TCL_OK;
}

int SetFrameArrowMag(ClientData,Tcl_Interp *interp,
		  int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc!=2) {
    sprintf(buf,"SetFrameArrowMag must be called with"
	    " 1 argument: new_arrowmag"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  REAL4 new_arrowmag=Omf_atof(argv[1]);
  if(new_arrowmag<0.0) {
    sprintf(buf,"Arrowmag must be >= 0.0");
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  strcpy(buf,"OMF_PushCursor ; OMF_SoftUpdate");
  Tcl_Eval(interp,buf);    // Save current cursor & display "watch" cursor

  // Change arrowmag in display list
  int errcode=TCL_OK;
  if(myFrame.SetArrowMag(new_arrowmag)!=0) errcode=TCL_ERROR;

  strcpy(buf,"OMF_PopCursor");
  Tcl_Eval(interp,buf);    // Retrieve original cursor

  return errcode;
}

int WriteMesh(ClientData,Tcl_Interp *interp,
	      int argc,char** argv)
{
  Tcl_ResetResult(interp);
  if(argc<4 || argc>6) {
    sprintf(buf,"WriteMesh must be called with"
	    " 3-5 arguments: filename <ascii|binary4|binary8>"
	    " <regular|irregular> [title] [description]"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  int errcode=TCL_OK;

  strcpy(buf,"OMF_PushCursor ; OMF_SoftUpdate");
  Tcl_Eval(interp,buf);    // Save current cursor & display "watch" cursor

  OmfDataStyle ods=omfi_ascii;
  if(strcmp("binary4",argv[2])==0)      ods=omfi_bin4;
  else if(strcmp("binary8",argv[2])==0) ods=omfi_bin8;

  BOOL force_irreg=0;
  if(strcmp("irregular",argv[3])==0) force_irreg=1;

  const char *title="",*desc="";
  if(argc>4) title=argv[4];
  if(argc>5) desc=argv[5];

  // Open file
  FILE *fptr=Omf_FOpen(argv[1],"wb");
  if(fptr==NULL) {
    sprintf(buf,"Error in WriteMesh: Unable to open file %s for output",
	    argv[1]);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    errcode=TCL_ERROR;
  }
  else {
    OmfFileOutput ofo;
    if(ofo.WriteMesh(myMesh,fptr,ods,force_irreg,title,desc)!=0) {
      sprintf(buf,"WriteMesh error");
      Tcl_AppendResult(interp,buf,(char *)NULL);
      errcode=TCL_ERROR;
    }
    fclose(fptr);
  }

  strcpy(buf,"OMF_PopCursor");
  Tcl_Eval(interp,buf);    // Retrieve original cursor

  return errcode;
}

int RegisterTclCommand(Tcl_Interp* interp,
			const char* name,Tcl_CmdProc* cmd)
{
  // I don't know that it is necessary to copy name to a
  // writeable buffer, but better safe than sorry.
  strncpy(buf,name,sizeof(buf));  buf[sizeof(buf)-1]='\0';
  if(Tcl_CreateCommand(interp,buf,cmd,NULL,NULL)==NULL) {
    fprintf(stderr,"Unable to register command ->%s<- "
	    "with Tcl Interpreter\n",buf);
    fflush(stderr);
    exit(1);
  }
  return TCL_OK;
}

int Mmdisp_Init(Tcl_Interp *interp)
{
  // Save window token and display pointer
  myTkWin=Tk_MainWindow(interp);
  myDisplay=Tk_Display(myTkWin);

  // Initialize C++ data structures
  myFrame.SetTclInterp(interp);
  myFrame.SetMesh(myMesh);
  myFrame.SetupFrame();

  // Register C/C++ routines with interpreter
  RegisterTclCommand(interp,"GetFirstFileName",GetFirstFileName);
  RegisterTclCommand(interp,"GetMeshType",GetMeshType);
  RegisterTclCommand(interp,"GetMeshName",GetMeshName);
  RegisterTclCommand(interp,"GetMeshTitle",GetMeshTitle);
  RegisterTclCommand(interp,"GetMeshDescription",GetMeshDescription);
  RegisterTclCommand(interp,"GetMeshIncrement",GetMeshIncrement);
  RegisterTclCommand(interp,"ChangeMesh",ChangeMesh);
  RegisterTclCommand(interp,"DrawFrame",DrawFrame);
  RegisterTclCommand(interp,"GetFrameRotation",GetFrameRotation);
  RegisterTclCommand(interp,"SetFrameRotation",SetFrameRotation);
  RegisterTclCommand(interp,"GetFrameSubsampling",GetFrameSubsampling);
  RegisterTclCommand(interp,"SetFrameSubsampling",SetFrameSubsampling);
  RegisterTclCommand(interp,"GetFrameSize",GetFrameSize);
  RegisterTclCommand(interp,"SetFrameSize",SetFrameSize);
  RegisterTclCommand(interp,"GetFrameScale",GetFrameScale);
  RegisterTclCommand(interp,"SetFrameScale",SetFrameScale);
  RegisterTclCommand(interp,"GetFrameArrowMag",GetFrameArrowMag);
  RegisterTclCommand(interp,"SetFrameArrowMag",SetFrameArrowMag);
  RegisterTclCommand(interp,"WriteMesh",WriteMesh);

#if SYSTEM_TYPE!=WINDOWS
  // Load icon bitmaps (doesn't work under Windows)
  Tk_DefineBitmap(interp,Tk_GetUid("omfIcon"),
		  omficon_bits,omficon_width,omficon_height);
  omfIconBitmap=Tk_GetBitmap(interp,myTkWin,Tk_GetUid("omfIcon"));
  Tk_DefineBitmap(interp,Tk_GetUid("omfIconMask"),
		  omfmask_bits,omfmask_width,omfmask_height);
  omfIconMask=Tk_GetBitmap(interp,myTkWin,Tk_GetUid("omfIconMask"));
#endif

// Enter mmdisp package into database and define standard globals and
// procs.  (Note: CVS will fill in the value for $Name:  $ below.  Need
// to configure CVS so that this Name and the one in the initialization
// script stay in sync.
//
// NOTE: Using "^" instead of "$" to delimit CVS keywords until we
//	set up the maintenance scripts in CVS.  This will prevent CVS
//	erasing the values we place there.  This also means we must
//	maintain these string by hand for now.
    strncpy(buf,"OMFCVS_PackageProvide {^Name: Mmdisp-0_0b0^} mmd",
	    sizeof(buf));
    if(Tcl_GlobalEval(interp,buf) != TCL_OK) return TCL_ERROR;

    // Pass the ServerFlag value to the Tcl interpreter
    sprintf(buf,"%d",ServerFlag);
    Tcl_SetVar(interp,"mmd_serverflag",buf,0);
    
    // Evaluates Mmdisp's initialization script
    strncpy(buf2,"OMF_SourceInitScript mmd",sizeof(buf2));
    if(Tcl_GlobalEval(interp,buf2) != TCL_OK) return TCL_ERROR;

    // Register cleanup routine
#if ((TCL_MAJOR_VERSION == 7) && (TCL_MINOR_VERSION >= 5)) \
    || (TCL_MAJOR_VERSION > 7)
    // Note: The following test would break against version number
    //   7.10, but Tcl is going from 7.6 to 8.0, so this shouldn't
    //   be a problem.
    if(Omf_atof(Tcl_GetVar(interp,"tcl_version",TCL_GLOBAL_ONLY))>7.4)
       Tcl_CreateExitHandler(TclCommandsCleanup,0);
    // We check dynamically so that dynamic builds against 7.5+
    //  libraries still have a chance to run using 7.4 libraries.
#endif

    return TCL_OK;
}

